/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import <Foundation/Foundation.h>
#import "IDLDefinition_Private.h"
#import "IDLTypedValue.h"
#import "IDLSequence.h"
#import "IDLArray.h"
#import "IDLDefinitionScope_Private.h"

@implementation IDLDefinition

static NSZone* _zone = nil;
static BOOL _allowsIndirection = YES;
static BOOL _allowsUndefinedKeys = NO;
static NSMutableSet *_errorKeys = nil;
static NSLock* _hashLock = nil;

void ORBLogUndefinedKeyError(IDLDefinition *def, id object, NSString* key)
{
	[_hashLock lock];
	if(!_errorKeys)
		_errorKeys = [[NSMutableSet alloc]init];

	NSString *className = NSStringFromClass([object class]);
	NSString *errorKey = [NSString stringWithFormat:@"$@-%@", className, key];
	if(![_errorKeys containsObject:errorKey]){
		[_errorKeys addObject:errorKey];
		NSLog(@"*** Class %@ is not KeyValueCoding compliant for key '%@', typeID '%@'. This error is reported only once.",
			  className, key, [def typeId]);
	}
	[_hashLock unlock];
}

+ (void) _registerDefaults
{
}

+ (void) initialize
{
    if([IDLDefinition class] == self){
        [[NSUserDefaults standardUserDefaults]registerDefaults:
            [NSDictionary dictionaryWithObjectsAndKeys:
				[NSNumber numberWithBool:_allowsIndirection], @"allowsIndirection",
				[NSNumber numberWithBool:_allowsUndefinedKeys], @"ORBAllowsUndefinedKeys", nil]];
        _zone = NSCreateZone(8192, 4096,NO);
        [self _registerDefaults];
        _allowsIndirection = [[NSUserDefaults standardUserDefaults]boolForKey:@"allowsIndirection"];
        _allowsUndefinedKeys = [[NSUserDefaults standardUserDefaults]boolForKey:@"ORBAllowsUndefinedKeys"];
		_hashLock = [[NSLock alloc]init];
    }
}

+ (NSZone*)zone
{
    return _zone;
}
- (NSZone*)zone
{
    return _zone;
}

- init
{
    self = [super init];
    _prefix = [[IDLDefinitionScope currentPrefix]retain];
    return self;
}

- (void) dealloc
{
    [_version release];
    [_prefix release];
    [_typeId release];

    [super dealloc];
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	if([coder allowsKeyedCoding]){
		[coder encodeObject:_typeId forKey:@"typeId"];
		[coder encodeObject:_prefix forKey:@"prefix"];
		[coder encodeObject:_version forKey:@"version"];
	}else{
		[coder encodeObject:_typeId];
		[coder encodeObject:_prefix];
		[coder encodeObject:_version];
	}
}

- initWithCoder:(NSCoder*)coder
{
	self = [super init];
	if([coder allowsKeyedCoding]){
		_typeId = [[coder decodeObjectForKey:@"typeId"]retain];
		_prefix = [[coder decodeObjectForKey:@"prefix"]retain];
		_version = [[coder decodeObjectForKey:@"version"]retain];
	}else{
		_typeId = [[coder decodeObject]retain];
		_prefix = [[coder decodeObject]retain];
		_version = [[coder decodeObject]retain];
	}
	return self;
}

- (id)awakeAfterUsingCoder:(NSCoder *)aDecoder
{
	id obj = [(NSObject*)[[IDLDefinitionScope currentScope]definitionForTypeId:_typeId]retain];
	if(obj && obj != self){
		[self release];
		return obj;
	}else
		return self;
}

+ definition
{
    self = [[[self allocWithZone:[self zone]]init] autorelease];
    return self;
}

- (NSString*) name
{
    return [NSString stringWithFormat:@"%x",self];
}


- (BOOL) isOneWay
{
    return NO;
}

- convertValue:aValue
{
    return aValue;
}

- coerceValue:aValue
{
    return aValue;
}

- (NSString*) valueDescriptionOf:aValue
{
    if([aValue isKindOfClass:[IDLTypedValue class]])
        return [(IDLTypedValue*)aValue valueDescription];
    else
        return [aValue description];
}

- sequenceClass
{
    return [IDLSequence class];
}

- arrayClass
{
    return [IDLArray class];
}

- (IDLDefinition*)referenceType
{
    return self;
}

- (void) setNamePath:(NSString*)path
{
// Don't reset typeID if already set
    if(_typeId == nil){
        NSString* prefix = [self prefix];
        path = path?[NSString stringWithFormat:@"%@/%@",path,[self name]] : [self name];
        //[_typeId release];
        if([prefix length] > 0)
            _typeId = [[NSString stringWithFormat:@"IDL:%@/%@:%@", prefix, path, [self version]]retain];
        else
            _typeId = [[NSString stringWithFormat:@"IDL:%@:%@", path, [self version]]retain];
    }
}

- (void) setVersion:(NSString*)aString
{
    if(_version == nil){
        //[_version release];
        _version = [aString copy];
    }
}

- (NSString*)version
{
    return _version?_version:@"1.0";
}

- (long) kind // TCKind
{
    return tk_null;
}

- (NSString*) prefix
{
    return _prefix;
}

- (NSString*) typeId
{
    return _typeId;
}

extern NSString* VXVersionFromTypeId(NSString* typeId);

- (void) setTypeId:(NSString*)str
{
    if(str != _typeId){
        [_typeId release];
        _typeId = [str retain];
        if(_version)
            [_version release];
        _version = [VXVersionFromTypeId(_typeId) retain];
    }
}

- (BOOL) isEqual:(IDLDefinition*)def
{
    if(self == def)
        return YES;
    return def && [def isMemberOfClass:[self class]] && def->_typeId && [_typeId isEqualToString:def->_typeId];
}

- (BOOL) _isDerivedFromIDLType:(IDLDefinition*)type
{
	return type == self;
}

- (BOOL) isDerivedFromIDLType:(IDLDefinition*)type
{
	IDLDefinition* refType = [type referenceType];
	return [[self referenceType]_isDerivedFromIDLType:refType];
}

+ (BOOL) allowsIndirection
{
    return _allowsIndirection;
}

+ (void) setAllowsIndirection:(BOOL)flag
{
    _allowsIndirection = flag?YES:NO;
}

+ (BOOL) allowsUndefinedKeys
{
	return _allowsUndefinedKeys;
}

+ (void) setAllowsUndefinedKeys:(BOOL)flag
{
	_allowsUndefinedKeys = flag? YES : NO;
}

@end
