/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLStream.h"
#import <Foundation/Foundation.h>
#import <Foundation/NSByteOrder.h>
#import "IDL.h"
#import "VXConnection.h"
#import "VXMessage.h"
#import "IDLGIOPVersion.h"

@implementation IDLStream

BOOL VXDecodeEmptyStringAsNil = YES; // If YES a string of length 0 is un-marshaled as nil
BOOL VXEncodeNilStringAsEmpty = YES; // If NO the 'nil' string is marshaled with length = 0
// The string with length = 0 always is un-marshaled to 'nil'
NSStringEncoding VXWStringTransmissionEncoding = NSUnicodeStringEncoding;
NSStringEncoding VXStringTransmissionEncoding = NSUTF8StringEncoding;

- (NSStringEncoding) wstringEncoding
{
	return _wstringEncoding;
}

- (NSStringEncoding) stringEncoding
{
	return _stringEncoding;
}

- (void) setStringEncoding:(NSStringEncoding)encoding
{
	_stringEncoding = encoding;
}

- (void) setWStringEncoding:(NSStringEncoding)encoding
{
	_wstringEncoding = encoding;
}


- initWithOffset:(unsigned)offset
{
    self = [self initWithStreamData:[NSMutableData data] offset:offset];
	_stringEncoding = VXStringTransmissionEncoding;
	_wstringEncoding = VXWStringTransmissionEncoding;
    _flags._isOutStream = YES;
    return self;
}

- init
{
    return [self initWithOffset:0];
}

- initWithStreamData:(NSMutableData*)data offset:(unsigned)offset
{
    self = [super init];
    _data = [data retain];
    _cursor = 0;
    _flags._offset = offset;
	_stringEncoding = VXStringTransmissionEncoding;
	_wstringEncoding = VXWStringTransmissionEncoding;
    return self;
}

- initWithStreamData:(NSMutableData*)data
{
    return [self initWithStreamData:data offset:0];
}

- (BOOL) isInputStream
{
    return !_flags._isOutStream;
}

- (void) dealloc
{
    [_data release];
    [super dealloc];
}

+ octetStreamWithData:(NSMutableData*)data offset:(unsigned)offset
{
    return [[[self alloc]initWithStreamData:data offset:offset]autorelease];
}

+ octetStreamWithData:(NSMutableData*)data
{
    return [self octetStreamWithData:data offset:0];
}
+ inputStreamWithOffset:(unsigned)offset
{
    return [self octetStreamWithData:[NSMutableData data] offset:offset];
}

+ inputStream
{
    return [self inputStreamWithOffset:0];
}

+ octetStreamWithOffset:(unsigned)offset
{
    return [[[self alloc]initWithOffset:offset]autorelease];
}

+ octetStream
{
    return [self octetStreamWithOffset:0];
}

- (void) useByteOrderFromFirstOctet
{
    if(!_flags._byteOrderInFirstOctet){ //don't do this twice
        _flags._byteOrderInFirstOctet = YES;
        if(_flags._isOutStream){
            unsigned char b = 0x00;
            switch(NSHostByteOrder()){
                case NS_LittleEndian:
                    b = 0x01;
                case NS_BigEndian:
                default:
                    [_data appendBytes:&b length:1];
            }
        }else{
            unsigned const char* b = [_data bytes];
            switch(NSHostByteOrder()){
                case NS_LittleEndian:
                    _flags._shouldSwap = ((b[0]&0x01) != 0x01);
                    break;
                case NS_BigEndian:
                default:
                    _flags._shouldSwap = ((b[0]&0x01) != 0x00);
                    break;
            }
            _cursor = 1;
        }
    }
}

// Returns whether it will swap
- (BOOL) setByteOrder:(unsigned)b // 0 - BigEndian, 1 - LittleEndian
{
    if(!_flags._isOutStream){
        switch(NSHostByteOrder()){
            case NS_LittleEndian:
                _flags._shouldSwap = ((b&0x01) != 0x01);
                break;
            case NS_BigEndian:
            default:
                _flags._shouldSwap = ((b&0x01) != 0x00);
                break;
        }
        return _flags._shouldSwap;
    }else
        return NO;
}

- (void) setIIOPVersion:(IDLGIOPVersion*)version
{
    _flags._major = [version majorVersion];
    _flags._minor = [version minorVersion];
}

- (void) reset
{
    _cursor = _flags._byteOrderInFirstOctet?1:0;
}

- (void) _alignTo:(unsigned int)b
{
    if(_flags._needsAlign){
        b = 8;
        _flags._needsAlign = NO;
    }
    if(b > 1){
        int length = [_data length] + _flags._offset;
        int l = length + b - 1 - (length+b-1)%b;
        if(l > length)
            [_data increaseLengthBy:(l-length)];
    }
}

- (void) _alignCursorTo:(unsigned int)b
{
    if(_flags._needsAlign){ // This cannot be inside chunk
        b = 8;
        _flags._needsAlign = NO;
    }
    
    if(b > 1){
        long l = _cursor - (_cursor + _flags._offset)%b;
        if(l < _cursor)
            l += b;
        _cursor = l;
        if(_cursor >= [_data length])
            [NSException raise:CORBAMarshalException format:@"cursor is beyond the boundary of stream"];
    }
}

- (BOOL) isAtEOF:(unsigned int)b
{
    long l = _cursor;

    if(_flags._needsAlign){ // This cannot be inside chunk
        b = 8;
    }
    
    if(b > 1){
        l = _cursor - (_cursor + _flags._offset)%b;
        if(l < _cursor)
            l += b;
    }
    if(l >= [_data length])
        return YES;
    else
        return NO;
}

- (void) setNeedsAlign:(BOOL)flag
{
    _flags._needsAlign = flag?YES:NO;
}

- (void) encodeChar:(unsigned char)value
{
    unsigned char val = value;
    [self _alignTo:sizeof(unsigned char)];
    [_data appendBytes:&val length:sizeof(char)];
}
- (void) encodeShort:(short)value
{
    short val = value;
    [self _alignTo:sizeof(short)];
    [_data appendBytes:&val length:sizeof(short)];
}
- (void) encodeUShort:(unsigned short)value
{
    unsigned short val = value;
    [self _alignTo:sizeof(short)];
    [_data appendBytes:&val length:sizeof(short)];
}
- (void) encodeLong:(long)value
{
    long val = value;
    [self _alignTo:sizeof(long)];
    [_data appendBytes:&val length:sizeof(long)];
}
- (void) encodeLong:(long)value atOffset:(unsigned)offset
{
    long size = [_data length];
    if(offset + sizeof(long) <= size && offset%sizeof(long) == 0){
        NSRange r = {offset,sizeof(long)};
        long val = value;
        [_data replaceBytesInRange:r withBytes:&val];
    }
}
- (void) encodeULong:(unsigned long)value
{
    unsigned long val = value;
    [self _alignTo:sizeof(long)];
    [_data appendBytes:&val length:sizeof(long)];
}
- (void) encodeBOOL:(BOOL)value
{
    BOOL val = value;
    [self _alignTo:sizeof(BOOL)];
    [_data appendBytes:&val length:sizeof(BOOL)];
}
- (void) encodeFloat:(float)value
{
    float val = value;
    [self _alignTo:sizeof(float)];
    [_data appendBytes:&val length:sizeof(float)];
}
- (void) encodeDouble:(double)value
{
    double val = value;
    [self _alignTo:sizeof(double)];
    [_data appendBytes:&val length:sizeof(double)];
}
- (void) encodeLongLong:(long long)value
{
    long long val = value;
    [self _alignTo:sizeof(long long)];
    [_data appendBytes:&val length:sizeof(long long)];
}
- (void) encodeULongLong:(unsigned long long)value
{
    unsigned long long val = value;
    [self _alignTo:sizeof(long long)];
    [_data appendBytes:&val length:sizeof(long long)];
}
- (void) encodeString:(NSString*)string
{
    if(string){ // encode <nil> as string with length 0
		NSData* bytes = [string dataUsingEncoding:[self stringEncoding] allowLossyConversion:YES];
        long len = [bytes length] + 1;
		char zero = 0;
        [self _alignTo:sizeof(long)];
        [_data appendBytes:&len length:sizeof(long)];
        [_data appendData:bytes];
		[_data appendBytes:&zero length:1];
    }else{
        if(VXEncodeNilStringAsEmpty){
            long len = 1;
            char zero = 0;
            [self _alignTo:sizeof(long)];
            [_data appendBytes:&len length:sizeof(long)];
            [_data appendBytes:&zero length:len];
        }else{
            long len = 0;
            [self _alignTo:sizeof(long)];
            [_data appendBytes:&len length:sizeof(long)];
        }
    }
}

- (void) encodeWString:(NSString*)string
{
// Not tested if wstring encoding is IIOP compliant
    if(string){ // encode <nil> as string with length 0
        NSData* bytes = [string dataUsingEncoding:[self wstringEncoding] allowLossyConversion:YES];
        long len = [bytes length];
        [self _alignTo:sizeof(long)];
        [_data appendBytes:&len length:sizeof(long)];
        [_data appendData:bytes];
    }else{
        long len = 0;
        [self _alignTo:sizeof(long)];
        [_data appendBytes:&len length:sizeof(long)];
    }
}

- (void) encodeUniChar:(unichar)value
{
// Not tested if IIOP compliant: wchar encoding 
    unichar val = value;
    if(_flags._minor >= 2){
        NSString* str = [NSString stringWithCharacters:&val length:1];
        NSData* data = [str dataUsingEncoding:[self wstringEncoding] allowLossyConversion:YES];
        unsigned char l = [data length];
        [self _alignTo:sizeof(unsigned char)];
        [_data appendBytes:&l length:1];
        [_data appendData:data];
    }else{
        [self _alignTo:sizeof(unichar)];
        [_data appendBytes:&val length:sizeof(unichar)];
    }
}

- (void) encodeObject:object
{
    IDLDefinition* iorObj = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:@"IOP/IOR"];
    if(iorObj){
        [self encodeStruct:object usingObjectCoder:iorObj];
    }else
        [NSException raise:CORBADefinitionException
                    format:@"Cannot find definition for 'IOP/IOR'"];
}
- (void) encodeStruct:object usingObjectCoder:objCoder
{
    [objCoder encodeObject:object withCoder:self];
}
- (void) encodeSequence:object usingObjectCoder:objCoder
{
	[objCoder encodeObject:object withCoder:self];
}
- (void) encodeData:stream
{
    long length = [(NSData*)stream length];
    [self _alignTo:sizeof(long)];
    [_data appendBytes:&length length:sizeof(long)];
    [_data appendData:stream];
}
- (void) encodeData:stream length:(unsigned)length
{
    NSRange r = {0,length};
    [self _alignTo:sizeof(char)];
    [_data appendData:[stream subdataWithRange:r]];
}

- (void) encodeRawData:(NSData*)stream
{
    [self _alignTo:sizeof(char)];
    [_data appendData:stream];
}

- (void) _swapBytes:(unsigned char*)buffer length:(unsigned)length
{
    register unsigned char r;
    register int n = 0, m = length-1;
    while(n < m){
        r = buffer[n]; buffer[n] = buffer[m]; buffer[m] = r;
        n += 1; m -=1;
    }
}

- (void) decodeChar:(unsigned char*)value
{
    const unsigned char* bytes = [_data bytes];
    [self _alignCursorTo:sizeof(unsigned char)];
    *value = bytes[_cursor++];
}
- (void) decodeShort:(short*)value
{
    NSRange r = {_cursor,sizeof(short)};
    [self _alignCursorTo:sizeof(short)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(short)];
}
- (void) decodeUShort:(unsigned short*)value
{
    NSRange r = {_cursor,sizeof(short)};
    [self _alignCursorTo:sizeof(short)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(short)];
}
- (void) decodeLong:(long*)value
{
    NSRange r = {_cursor,sizeof(long)};
    [self _alignCursorTo:sizeof(long)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(long)];
}

- (void) decodeULong:(unsigned long*)value
{
    NSRange r = {_cursor,sizeof(long)};
    [self _alignCursorTo:sizeof(long)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(long)];
}
- (void) decodeBOOL:(BOOL*)value
{
    NSRange r = {_cursor,sizeof(BOOL)};
    [self _alignCursorTo:sizeof(BOOL)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
}
- (void) decodeFloat:(float*)value
{
    NSRange r = {_cursor,sizeof(float)};
    [self _alignCursorTo:sizeof(float)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(float)];
}
- (void) decodeDouble:(double*)value
{
    NSRange r = {_cursor,sizeof(double)};
    [self _alignCursorTo:sizeof(double)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(double)];
}
- (void) decodeLongLong:(long long*)value
{
    NSRange r = {_cursor,sizeof(long long)};
    [self _alignCursorTo:sizeof(long long)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(long long)];
}
- (void) decodeULongLong:(unsigned long long*)value
{
    NSRange r = {_cursor,sizeof(long long)};
    [self _alignCursorTo:sizeof(long long)];
    r.location = _cursor;
    [_data getBytes:value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)value length:sizeof(long long)];
}


- (void) decodeString:(NSString**)value
{
    NSRange r = {_cursor,sizeof(long)};
    long length;
    [self _alignCursorTo:sizeof(long)];
    r.location = _cursor;
    [_data getBytes:&length range:r];
    r.location += r.length;
    _cursor = r.location;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)&length length:sizeof(long)];

    if(r.location + length > [_data length])
        [NSException raise:CORBAMarshalException 
                format:@"string length beyond buffer"];
	
	_cursor += length;
    if(length > 1){ // length = 0 ==> <nil>
		r.length = length - 1; // Do not include terminating 0
		NSData* bytes = [_data subdataWithRange:r];
		*value = [[[NSString allocWithZone:[self zone]]initWithData:bytes encoding:[self stringEncoding]]autorelease];
    }else
	if(length == 1)
		*value = VXDecodeEmptyStringAsNil? nil : @"";
	else
        *value = nil;
}

- (void) decodeWString:(NSString**)value
{
// Not tested if IIOP compliant: wstring decoding
    NSRange r = {_cursor,sizeof(long)};
    long length;
    [self _alignCursorTo:sizeof(long)];
    r.location = _cursor;
    [_data getBytes:&length range:r];
    r.location += r.length;
    _cursor = r.location;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)&length length:sizeof(long)];
    if(length > 0){ // length = 0 ==> <nil>
        NSData* bytes;
        r.length = length;
        bytes = [_data subdataWithRange:r];
        _cursor = r.location + r.length;
        
        *value = [[[NSString allocWithZone:[self zone]]initWithData:bytes encoding:[self wstringEncoding]]autorelease];
    }else{
        if(VXDecodeEmptyStringAsNil)
            *value = nil;
        else
            *value = @"";
    }
}

- (void) decodeUniChar:(unichar*)value
{
    if(_flags._minor >= 2){
        const unsigned char* bytes = [_data bytes];
        unsigned char l;
        NSRange r;
        NSData* data;
        NSString* str;
        [self _alignCursorTo:sizeof(unsigned char)];
        l = bytes[_cursor++];
        r.location = _cursor;
        r.length = l;
        data = [_data subdataWithRange:r];
        str = [[[NSString allocWithZone:[self zone]]initWithData:data encoding:[self wstringEncoding]]autorelease];
        *value = [str characterAtIndex:0];
        _cursor += r.length;
    }else{
        NSRange r = {_cursor,sizeof(unichar)};
        [self _alignCursorTo:sizeof(unichar)];
        r.location = _cursor;
        [_data getBytes:value range:r];
        _cursor += r.length;
        if(_flags._shouldSwap)
            [self _swapBytes:(unsigned char*)value length:sizeof(unichar)];
    }
}

- (void) decodeObject:(id*)object usingClass:objectClass
{
    IDLDefinition* iorObj = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:@"IOP/IOR"];
    if(iorObj){
        [self decodeStruct:object usingObjectCoder:iorObj objectClass:objectClass];
    }else
        [NSException raise:CORBADefinitionException
                    format:@"Cannot find definition for 'IOP/IOR'"];
}

- (void) decodeObject:(id*)object
{
    IDLDefinition* iorObj = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:@"IOP/IOR"];
    if(iorObj){
        [self decodeStruct:object usingObjectCoder:iorObj];
    }else
        [NSException raise:CORBADefinitionException
                    format:@"Cannot find definition for 'IOP/IOR'"];
}
- (void) peepObject:(id*)object
{
    IDLDefinition* iorObj = (IDLDefinition*)[[IDLDefinitionScope rootScope]definitionForNamePath:@"IOP/IOR"];
    if(iorObj){
        int cursor = _cursor;
        [self decodeStruct:object usingObjectCoder:iorObj];
        _cursor = cursor; // restore cursor
    }else
        [NSException raise:CORBADefinitionException
                    format:@"Cannot find definition for 'IOP/IOR'"];
}

- (void) decodeStruct:(id*)object usingObjectCoder:coder
{
    [coder decodeObject:object withCoder:self];
}
- (void) decodeStruct:(id*)object usingObjectCoder:coder objectClass:objectClass
{
    [coder decodeObject:object withCoder:self usingClass:objectClass];
}
- (void) peepStruct:(id*)object usingObjectCoder:coder
{
    int cursor = _cursor;
    [coder decodeObject:object withCoder:self];
    _cursor = cursor;
}

- (void) decodeSequence:(id*)object usingObjectCoder:coder
{
    NSRange r = {_cursor,sizeof(long)};
    long count;
    [self _alignCursorTo:sizeof(long)];
    r.location = _cursor;
    [_data getBytes:&count range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)&count length:sizeof(long)];
    *object = [NSMutableArray array];
    while(count-- > 0){
        id item = nil;
        [coder decodeObject:&item withCoder:self];
        if(item)
            [*object addObject:item];
        else
            [*object addObject:[NSNull null]];
    }
}

- (void) decodeStringSequence:(id*)object
{
    [self doesNotRecognizeSelector:_cmd];
}

- (void) decodeData:(id*)stream
{
    NSRange r = {_cursor,sizeof(long)};
    long length;
    [self _alignCursorTo:sizeof(long)];
    r.location = _cursor;
    [_data getBytes:&length range:r];
    r.location += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)&length length:sizeof(long)];
    r.length = length;
    *stream = [NSData dataWithData:[_data subdataWithRange:r]];
    _cursor = r.location + r.length;
}
- (void) decodeData:(id*)stream length:(unsigned)length
{
    NSRange r = {_cursor,length};
    [self _alignCursorTo:sizeof(char)];
    r.location = _cursor;
    *stream = [NSData dataWithData:[_data subdataWithRange:r]];
    _cursor += r.length;
}

- (BOOL) decodeRawData:(NSData**)stream length:(long)length
{
    [self _alignCursorTo:sizeof(char)];
    if(_cursor + length <= [_data length]){
        NSRange r = {_cursor, length};
        *stream = [_data subdataWithRange:r];
        _cursor += length;
        return YES;
    }else{
        *stream = nil;
        return NO;
    }
}

- (unsigned) length
{
    return [_data length];
}

- (long) nextLocationForSize:(unsigned)size
{
    long l = _flags._isOutStream?[_data length] : _cursor;
    if(_flags._needsAlign)
        size = 8;
    if(size > 1){
        long l1 = l - (l+_flags._offset)%size;
        if(l1 < l)
            l = l1 + size;
        else
            l = l1;
    }
    return l;
}

- (long) location // for sizeof(long)
{
    return [self nextLocationForSize:sizeof(long)];
}

- (unsigned) cursorPosition
{
	return _cursor;
}
- (void)setCursorPosition:(unsigned)position
{
	_cursor = position;
}

- (NSData*) data
{
    return _data;
}

- (void) encodeIndirection:(long)location
{
    [self doesNotRecognizeSelector:_cmd];
}
- (IDLDefinition*) definitionWithLocation:(long)location
{
    [self doesNotRecognizeSelector:_cmd];
    return nil;
}
- (void) registerDefinition:(IDLDefinition*)def forLocation:(long)loc
{
    [self doesNotRecognizeSelector:_cmd];
}

- (long) locationForDefinition:(IDLDefinition*)def
{
    return 0;
}

- (void) registerObject:object withLocation:(long)location
{
    [self doesNotRecognizeSelector:_cmd];
}

- objectWithLocation:(long)location
{
    [self doesNotRecognizeSelector:_cmd];
    return nil;
}

- (void) decodeValue:(id*)obj forObjectCoder:(IDLValueType*)valuetype
{
    [self doesNotRecognizeSelector:_cmd];
}

- (void) encodeValue:obj forObjectCoder:(IDLValueType*)valuetype
{
    [self doesNotRecognizeSelector:_cmd];
}

- (VXMessage*)message
{
	[self doesNotRecognizeSelector:_cmd];
	return nil;
}

- (void) setMessage:(VXMessage*)message
{
	[self doesNotRecognizeSelector:_cmd];
}

- invocationObject
{
	return [[self message]invocationObject];
}

@end
